@@ -17,7 +17,7 @@
  *          Matthew Saltzman, Clemson University                             *
  *                                                                           * 
  *                                                                           *
- * Copyright (C) 2001-2007, Lehigh University, Yan Xu, and Ted Ralphs.       *
+ * Copyright (C) 2001-2008, Lehigh University, Yan Xu, and Ted Ralphs.       *
  *===========================================================================*/
 
 #include <cmath>
@@ -120,6 +120,7 @@ AlpsSubTree::AlpsSubTree()
     nodePool_(new AlpsNodePool), 
     diveNodePool_(new AlpsNodePool), 
     diveNodeRule_(new AlpsNodeSelectionEstimate),
+    diveDepth_(0),
     activeNode_(0),
     quality_(ALPS_OBJ_MAX),
     broker_(0)
@@ -139,6 +140,7 @@ AlpsSubTree::AlpsSubTree(AlpsKnowledgeBroker* kb)
     nodePool_(new AlpsNodePool),
     diveNodePool_(new AlpsNodePool),
     diveNodeRule_(new AlpsNodeSelectionEstimate),
+    diveDepth_(0),
     activeNode_(0),
     quality_(ALPS_OBJ_MAX)
 { 
@@ -325,7 +327,14 @@ AlpsSubTree::calculateQuality()
     int nodeSelectionType = broker_->getNodeSelection()->getType();
     const int nodeNum = nodePool_->getNumKnowledges();
     const int diveNum = diveNodePool_->getNumKnowledges();
-    
+
+    // Check if no node.
+    if ( (nodeNum + diveNum <= 0) && (activeNode_ == NULL) ) {
+	std::cout << "PROC[" << getKnowledgeBroker()->getProcRank()
+		  << "] has a subtree with no node" << std::endl;
+	assert(0);
+    }
+
     if ( ((nodeSelectionType == AlpsSearchTypeBestFirst) ||
 	  (nodeSelectionType == AlpsSearchTypeHybrid)) &&
 	 (eliteSize == 1) ) {
@@ -342,26 +351,28 @@ AlpsSubTree::calculateQuality()
 	return quality_;
     }
     
-    if (nodeNum <= 0) {
-	std::cout << "PROC[" << getKnowledgeBroker()->getProcRank()
-		  << "] has a subtree with no node" << std::endl;
-	assert(nodeNum > 0);
+    // Diving pool has no nodes if not using hybrid search.
+    assert(diveNum == 0);
+
+    if (activeNode_) {
+        if ( (activeNode_->getStatus() != AlpsNodeStatusFathomed) &&
+             (activeNode_->getStatus() != AlpsNodeStatusBranched) ) {
+            quality_ = activeNode_->getQuality();
+        }
     }
-   
-    std::vector<AlpsTreeNode* > allNodes = 
+
+    // Get best quality values of nodes in node pool.
+    std::vector<AlpsTreeNode* > nodeVector = 
 	nodePool_->getCandidateList().getContainer();
-    const int numNodes = getNumNodes();
-    std::vector<AlpsTreeNode* >::iterator pos = allNodes.begin();
+    std::vector<AlpsTreeNode* >::iterator pos = nodeVector.begin();
      
     std::multimap<double, AlpsTreeNode*> eliteList;
     std::multimap<double, AlpsTreeNode*>::iterator posEnd;    
-
-    for (int i = 0; i < numNodes; ++i) {
+    
+    for (int i = 0; i < nodeNum; ++i) {
 	AlpsTreeNode* node = (*pos);
 	double quality = (*pos)->getQuality();
-
 	++pos;
-        
         if (eliteSize == 1) {
             if (quality_ > quality) {
                 quality_ = quality;
@@ -369,15 +380,16 @@ AlpsSubTree::calculateQuality()
         }
         else {
             if (static_cast<int>(eliteList.size()) < eliteSize) {
-                eliteList.insert(std::pair<double,AlpsTreeNode*>(quality,node));
+                std::pair<const double,AlpsTreeNode*> sa(quality,node);
+                eliteList.insert(sa);
 		
             }
             else {  // ==   
                 posEnd = eliteList.end();
                 --posEnd;
                 if (quality < posEnd->first) {
-                    eliteList.insert(std::pair<double, AlpsTreeNode*>
-				 (quality, node));
+                    std::pair<const double, AlpsTreeNode*> sa(quality, node);
+                    eliteList.insert(sa);
                     posEnd = eliteList.end();
                     --posEnd;
                     eliteList.erase(posEnd);
@@ -633,10 +645,35 @@ AlpsSubTree::splitSubTree(int& returnSize, int size)
     AlpsTreeNode* subTreeRoot = 0;
     AlpsTreeNode* rootParent = 0;
 
+    int nodeMemSize = getKnowledgeBroker()->getNodeMemSize();
     int LS = broker_->getLargeSize()/2;
-    int maxAllowNodes = LS / getKnowledgeBroker()->getNodeMemSize();
-    maxAllowNodes = (maxAllowNodes > 0) ? maxAllowNodes : 1;
+    int maxAllowNodes = LS / nodeMemSize;
     
+    if (maxAllowNodes == 0) {
+        returnSize = 0;
+        return st;
+    }
+
+    // At most send 50 nodes
+    if (nodeMemSize < 10000) {
+        maxAllowNodes = (maxAllowNodes > 100) ? 100 : maxAllowNodes;
+    }
+    else if (nodeMemSize < 50000) {
+        maxAllowNodes = (maxAllowNodes > 50) ? 30 : maxAllowNodes;
+    }
+    else if (nodeMemSize < 100000) {
+        maxAllowNodes = (maxAllowNodes > 30) ? 20 : maxAllowNodes;
+    }
+    else if (nodeMemSize < 500000) {
+        maxAllowNodes = (maxAllowNodes > 10) ? 5 : maxAllowNodes;
+    }
+    else if (nodeMemSize < 1000000) {
+        maxAllowNodes = (maxAllowNodes > 3) ? 3 : maxAllowNodes;
+    }
+    else {
+        maxAllowNodes = 1;
+    }
+
 #if 0
     //------------------------------------------------------
     // This is a way to find the subtree root. Do a breath first search
@@ -868,7 +905,7 @@ AlpsSubTree::encode() const
     // root of the subtree.
     //------------------------------------------------------
 
-    std::vector<AlpsTreeNode* > allNodes;
+    std::vector<AlpsTreeNode* > nodeVector;
     std::stack<AlpsTreeNode* > nodeStack;
 
     int i = -1, nodeNum = 0; 
@@ -881,7 +918,7 @@ AlpsSubTree::encode() const
     while( !nodeStack.empty() ) {
 	curNode = nodeStack.top();
 	nodeStack.pop();
-	allNodes.push_back(curNode);         // The first is root_ 
+	nodeVector.push_back(curNode);         // The first is root_ 
  
 	numChildren = curNode->getNumChildren();
 	for (i = 0; i < numChildren; ++i) {
@@ -889,7 +926,7 @@ AlpsSubTree::encode() const
 	}
     }
 
-    nodeNum = static_cast<int> (allNodes.size());
+    nodeNum = static_cast<int> (nodeVector.size());
 
     AlpsEncoded* encoded = new AlpsEncoded(AlpsKnowledgeTypeSubTree);
 
@@ -898,7 +935,7 @@ AlpsSubTree::encode() const
     AlpsTreeNode* node = NULL;
   
     for (i = 0; i < nodeNum; ++i) {          // Write all nodes to rep of enc
-	node = allNodes[i];
+	node = nodeVector[i];
 	encoded->writeRep(node->getExplicit()); 
 	AlpsEncoded* enc = node->encode();
 	encoded->writeRep(enc->size());
@@ -1068,7 +1105,6 @@ AlpsSubTree::decode(AlpsEncoded& encoded) const
     return st;
 }
 
-#if 1
 //#############################################################################
 
 AlpsReturnStatus
@@ -1102,8 +1138,10 @@ AlpsSubTree::exploreUnitWork(bool leaveAsIt,
 
     AlpsSearchStrategy<AlpsTreeNode*> *nodeSel = broker_->getNodeSelection();
 
+#ifdef ALPS_MEMORY_USAGE
     bool checkMemory = broker_->getModel()->AlpsPar()->
 	entry(AlpsParams::checkMemory);
+#endif
     
     //------------------------------------------------------
     // Check if required to exit when find a solution.
@@ -1189,8 +1227,8 @@ AlpsSubTree::exploreUnitWork(bool leaveAsIt,
 		  //  << std::endl;
 	       }
 	    }
-
-            // Print node log.
+            
+            // Print node log for serial code.
             broker_->getModel()->nodeLog(activeNode_, forceLog);
 	    forceLog = false;
 
@@ -1281,237 +1319,6 @@ AlpsSubTree::exploreUnitWork(bool leaveAsIt,
     return status;
 }
 
-#else
-//#############################################################################
-
-AlpsReturnStatus
-AlpsSubTree::exploreUnitWork(int unitWork,
-                             double unitTime,
-                             AlpsExitStatus & exploreStatus, /* Output */
-                             int & numNodesProcessed,       /* Output */
-                             int & depth,                   /* Output */
-                             bool & betterSolution)         /* Output */
-{
-    // Start to count time.
-    broker_->subTreeTimer().start();
-
-    AlpsReturnStatus status = AlpsReturnStatusOk;
-    
-    int numChildren = 0;
-        
-    double oldSolQuality = ALPS_OBJ_MAX;
-    double newSolQuality = ALPS_OBJ_MAX;
-
-    AlpsTreeNode * tempNode = NULL;
-
-    bool forceLog = false;
-    
-    //------------------------------------------------------    
-    // Get parameters.
-    //------------------------------------------------------
-
-    const bool deleteNode = 
-	broker_->getModel()->AlpsPar()->entry(AlpsParams::deleteDeadNode);
-    
-    //------------------------------------------------------
-    // If need to stop whenever find a better solution.
-    //------------------------------------------------------
-
-    bool exitIfBetter = false;
-    if (betterSolution) {
-        exitIfBetter = true;
-        betterSolution = false;
-    }
-    
-    if( broker_->hasKnowledge(AlpsKnowledgeTypeSolution) ) {
-        oldSolQuality =
-	   broker_->getBestKnowledge(AlpsKnowledgeTypeSolution).second;
-    }    
-
-    //------------------------------------------------------
-    // Process nodes until reach unit limits, or better solution if check.
-    //------------------------------------------------------    
-
-    activeNode_ = NULL;
-    exploreStatus = AlpsExitStatusInfeasible;    
-    numNodesProcessed = 0;
-
-    while ( (nodePool_->hasKnowledge() || activeNode_) && 
-	    !betterSolution ) {
-	
-	broker_->subTreeTimer().stop();
-	broker_->timer().stop();
-	
-	if (numNodesProcessed > unitWork) {
-            exploreStatus = AlpsExitStatusNodeLimit;
-	    break;
-	}
-	else if (broker_->subTreeTimer().getTime() > unitTime) {
-            exploreStatus = AlpsExitStatusTimeLimit;
-	    // getKnowledgeBroker()->setSolStatus(AlpsExitStatusTimeLimit);
-	    break;
-	}
-        
-	// Get next node to be processed.
-	if (!activeNode_) {
-            activeNode_ = dynamic_cast<AlpsTreeNode*>
-		(nodePool_->getKnowledge().first); 
-            activeNode_->setDiving(false);
-            
-#ifdef NF_DEBUG_MORE
-            std::cout << "======= NOTE[" << activeNode_->getIndex() 
-                      << "]: JUMP : depth = " << activeNode_->getDepth() 
-                      << ", quality = " << activeNode_->getQuality()
-                      << ", estimate = " << activeNode_->getSolEstimate()
-                      << std::endl;
-#endif
-            nodePool_->popKnowledge();
-	}
-        else {
-            activeNode_->setDiving(true);
-        }
-        
-	switch (activeNode_->getStatus()) {
-	case AlpsNodeStatusPregnant: 
-        {            
-	    if (depth < activeNode_->getDepth() + 1) {
-		// Record the depth of tree
-		depth = activeNode_->getDepth() + 1;
-            }
-            
-	    // Should be part of default node selection.
-
-            // (1) Move left nodes in diving pool to normal pool.
-            while (diveNodePool_->getNumKnowledges() > 0) {
-                tempNode = dynamic_cast<AlpsTreeNode *>
-                    (diveNodePool_->getKnowledge().first);
-                diveNodePool_->popKnowledge();
-                nodePool_->addKnowledge(tempNode, tempNode->getQuality());
-            }
-
-            // (2) Branching to form children descriptions.
-	    std::vector< CoinTriple<AlpsNodeDesc*, AlpsNodeStatus, double> > 
-		children = activeNode_->branch();
-            
-	    // (3) Create actual children nodes and store in diving pool
-            createChildren(activeNode_, children, diveNodePool_);
-            numChildren = diveNodePool_->getNumKnowledges();
-            
-            // (4) Get a new node for diving pool.
-            if (numChildren > 0) {
-                activeNode_ = dynamic_cast<AlpsTreeNode *>
-                    (diveNodePool_->getKnowledge().first);
-                diveNodePool_->popKnowledge();
-            }
-            else {
-                activeNode_ = NULL;
-            }
-	    break;
-	}
-	case AlpsNodeStatusCandidate:
-	case AlpsNodeStatusEvaluated:
-	    activeNode_->setActive(true);
-
-            // Process the node.
-	    if (activeNode_ == root_) {
-                activeNode_->process(true);
-	    }
-	    else {
-                activeNode_->process();
-	    }           
-	    activeNode_->setActive(false); 
-
-	    // Record the new sol quality if have.
-	    if( broker_->hasKnowledge(AlpsKnowledgeTypeSolution) ) {
-	       newSolQuality = 
-		  broker_->getBestKnowledge(AlpsKnowledgeTypeSolution).second;
-	       if (newSolQuality < oldSolQuality) {
-		  if (exitIfBetter) {
-		     betterSolution = true;
-		  }
-		  forceLog = true;
-		  exploreStatus = AlpsExitStatusFeasible;
-		  oldSolQuality = newSolQuality;
-		  
-		  // std::cout << "betterSolution value=" << newSolQuality
-		  //  << std::endl;
-	       }
-	    }
-
-            // Print node log.
-            broker_->getModel()->nodeLog(activeNode_, forceLog);
-	    forceLog = false;
-
-            // Increment by 1.
-	    ++numNodesProcessed;
-            
-            // Check processing status.
-	    switch (activeNode_->getStatus()) {
-	    case AlpsNodeStatusCandidate :
-	    case AlpsNodeStatusEvaluated :
-	    case AlpsNodeStatusPregnant :
-		break;
-	    case AlpsNodeStatusFathomed :
-		if (deleteNode) {
-		    removeDeadNodes(activeNode_);
-                }
-                if (diveNodePool_->getNumKnowledges() > 0) {
-                    activeNode_ = dynamic_cast<AlpsTreeNode *>
-                        (diveNodePool_->getKnowledge().first);
-                    diveNodePool_->popKnowledge();
-                }
-                else {
-                    activeNode_ = NULL;
-                }
-		break;
-	    default : 
-                // AlpsNodeStatus::branched ==> this is impossible
-		throw CoinError("Impossible status: branched", 
-				"exploreSubTree", "AlpsSubTree"); 
-	    }
-	    break;
-	default : // branched or fathomed
-	    throw CoinError("Impossible status: branched or fathomed", 
-			    "exploreSubTree", "AlpsSubTree"); 
-	}
-    }
-
-    //------------------------------------------------------
-    // Clear up.
-    //------------------------------------------------------
-
-    // Possible reasons for breaking while:
-    // 1 no nodes:        do nothing
-    // 2 reach limits:    move nodes from diving pool to regular pool
-    // 3 better solution: move nodes from diving pool to regular pool
-    
-    if ( (exploreStatus == AlpsExitStatusTimeLimit) ||
-         (exploreStatus == AlpsExitStatusNodeLimit) ||
-         (exploreStatus == AlpsExitStatusFeasible) ) {
-        // Case 2 and 3.
-
-        // Move nodes in diving pool to normal pool.
-        while (diveNodePool_->getNumKnowledges() > 0) {
-            tempNode = dynamic_cast<AlpsTreeNode *>
-                (diveNodePool_->getKnowledge().first);
-            diveNodePool_->popKnowledge();
-            nodePool_->addKnowledge(tempNode, tempNode->getQuality());
-        }
-        if (activeNode_) {
-            nodePool_->addKnowledge(activeNode_, activeNode_->getQuality());
-        }
-    }
-    else {
-        // case 1.
-        assert(nodePool_->getNumKnowledges() == 0);
-        assert(diveNodePool_->getNumKnowledges() == 0);
-        assert(activeNode_ == NULL);
-    }
-
-    return status;
-}
-#endif
-
 //#############################################################################
 
 double 
